# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import os
import pathlib
import threading

import flask
import werkzeug.serving

bp = flask.Blueprint('tuxgo-uploader', __name__)

@bp.route('/', methods=['GET'])
def index():
    message = ''.join(flask.get_flashed_messages())
    if message:
        flask.current_app.config['is_shutdown'].set()
    else:
        message = 'Pick image below (will open camera) and submit:'
        
    return f'''\
<!doctype html>
<meta name="viewport" content="width=device-width, initial-scale=1" />
<title>TuxGo capture image</title>
<body>
    <p>{message}</p>
    <form method="post" enctype="multipart/form-data">
        <p><input name="file" type="file" capture="environment" accept="image/*" /></p>
        <p><input type="submit"></p>
    </form>
</body>
'''

@bp.route('/', methods=['POST'])
def submit():
    file = flask.request.files['file']
    flask.current_app.config['image'] = file.read()
    flask.flash(f'the image was submitted')
#   filename = f'/tmp/download.{file.filename.rsplit(".", 1)[-1].casefold()}'
#   file.save(filename)
#   flask.flash(f'the image was saved at {filename}')
    return flask.redirect(flask.url_for('.index'), 303)

def create_app():
    app = flask.Flask(__name__)
    app.secret_key = os.urandom(16)
    app.config['MAX_CONTENT_LENGTH'] = int(16e6) # 16 MB
    app.config['is_shutdown'] = threading.Event()
    app.register_blueprint(bp)
    return app

def get_image_from_http_server(host='0.0.0.0', port=5000, **kwds):
    app = create_app()
    server = werkzeug.serving.make_server(host, port, app, **kwds)
    thread = threading.Thread(target=server.serve_forever)
    thread.start()
    app.config['is_shutdown'].wait()
    server.shutdown()
    thread.join()
    return app.config.get('image')

if __name__ == '__main__':
#   create_app().run('0.0.0.0')
    print(repr(get_image_from_http_server()[:100]))



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

"""
Engine might be either a visitor or a virtual machine
"""

import abc
import asyncio
import dataclasses
import enum
import importlib.resources
import itertools
import operator
import sys
import types
import typing

from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
)

from loguru import logger

from . import (
    ast,
    blocks,
    bluetooth,
)


class BaseEngine(metaclass=abc.ABCMeta):
    def __init__(self, functions=types.MappingProxyType({})):
        # pylint: disable=redefined-outer-name
        self.functions = {**functions}
        self.variables = {}

#       self.pc = None
#       self.block_markers = []
#       self.breakpoints = set()

    def exec_run(self):
        while not self.exec_step():
            pass

    def exec_step(self):
        return next(self.instructions)()

    #
    # The following methods need to be overloaded in the engine implementation:
    #

    @abstractmethod
    def eval_Number(self, number: ast.Number):
        raise NotImplementedError()

    @abstractmethod
    def eval_Variable(self, variable: ast.Variable):
        raise NotImplementedError()

    @abstractmethod
    def eval_BinOp(self, binop: ast.BinOp):
        raise NotImplementedError()

    @abstractmethod
    def eval_Condition(self, condition: ast.Condition):
        raise NotImplementedError()


    @abstractmethod
    def execute_Repeat(self, repeat: ast.Repeat):
        raise NotImplementedError()

    @abstractmethod
    def execute_RepeatWhile(self, repeat_while: ast.RepeatWhile):
        raise NotImplementedError()

    @abstractmethod
    def execute_Step(self, step: ast.Step):
        raise NotImplementedError()

    @abstractmethod
    def execute_TurnLeft(self, turn_left: ast.TurnLeft):
        raise NotImplementedError()

    @abstractmethod
    def execute_TurnRight(self, turn_right: ast.TurnRight):
        raise NotImplementedError()

    @abstractmethod
    def execute_Jump(self, jump: ast.Jump):
        raise NotImplementedError()

    @abstractmethod
    def execute_PickUp(self, pick_up: ast.PickUp):
        raise NotImplementedError()

    @abstractmethod
    def execute_Activate(self, activate: ast.Activate):
        raise NotImplementedError()

    @abstractmethod
    def execute_Draw(self, draw: ast.Draw):
        raise NotImplementedError()

    @abstractmethod
    def execute_Place(self, place: ast.Place):
        raise NotImplementedError()

    @abstractmethod
    def execute_CallFunction(self, call_function: ast.CallFunction):
        raise NotImplementedError()

    @abstractmethod
    def execute_If(self, if_: ast.If):
        raise NotImplementedError()

    @abstractmethod
    def execute_Else(self, else_: ast.Else):
        raise NotImplementedError()

    @abstractmethod
    def execute_Programme(self, programme: ast.Programme):
        raise NotImplementedError()

#class BaseVisitorEngine(BaseEngine):
#    @abstractmethod
#    def default_visit(self, node):
#        raise NotImplementedError()
#
#    def eval_Number(self, number: ast.Number):
#        self.visit_Number(number)
#    def visit_Number(self, number: ast.Number):
#        self.default_visit(number)
#
#    def eval_Variable(self, variable: ast.Variable):
#        self.visit_Variable(variable)
#    def visit_Variable(self, variable: ast.Variable):
#        self.default_visit(variable)
#
#    def eval_BinOp(self, binop: ast.BinOp):
#        self.visit_BinOp(binop)
#    def visit_BinOp(self, binop: ast.BinOp):
#        self.default_visit(binop)
#
#    def eval_Condition(self, condition: ast.Condition):
#        self.default_visit()
#
#    def execute_Repeat(self, repeat: ast.Repeat):
#        self.default_visit()
#
#    def execute_RepeatWhile(self, repeat_while: ast.RepeatWhile):
#        self.default_visit()
#
#    def execute_Step(self, step: ast.Step):
#        self.default_visit()
#
#    def execute_TurnLeft(self, turn_left: ast.TurnLeft):
#        self.default_visit()
#
#    def execute_TurnRight(self, turn_right: ast.TurnRight):
#        self.default_visit()
#
#    def execute_Jump(self, jump: ast.Jump):
#        self.default_visit()
#
#    def execute_PickUp(self, pick_up: ast.PickUp):
#        self.default_visit()
#
#    def execute_Activate(self, activate: ast.Activate):
#        self.default_visit()
#
#    def execute_Draw(self, draw: ast.Draw):
#        self.default_visit()
#
#    def execute_Place(self, place: ast.Place):
#        self.default_visit()
#
#    def execute_CallFunction(self, call_function: ast.CallFunction):
#        self.default_visit()
#
#    def execute_If(self, if_: ast.If):
#        self.default_visit()

# original ScottieGo forbids redefining variables, which are more like constants
class _ConstDict(dict):
    def __setitem__(self, key, value):
        if key in self:
            raise TypeError(f'cannot redefine variable {key}')
        super().__setitem__(key, value)

class ExecEngine(BaseEngine):
    class Break(BaseException):
        pass
    class BreakFunction(BaseException):
        pass

    def __init__(self, *args, allow_variable_reassignment=True, **kwds):
        super().__init__(*args, **kwds)
        self.variables = {} if allow_variable_reassignment else _ConstDict()
        self.functions = {}
        self.main: Optional[ast.Programme] = None

    def eval_Number(self, number):
        return number.value

    def eval_Variable(self, variable):
        try:
            return self.variables[variable.name]
        except KeyError:
            raise NameError(
                f'variable {self.varname} referenced before assignment '
                f'at line {self.line} block {self.cols[0]}')

    def eval_BinOp(self, binop):
        return binop.op(binop.left.eval(engine), binop.right.eval(engine))

    def eval_Condition(self, condition):
        return self.detect(condition.direction, condition.object)


    def execute_block(self, block):
        # TODO bracket
        for stmt in block:
            stmt.execute(self)

    def execute_Programme(self, programme):
        self.execute_block(programme.body)

    def execute_Repeat(self, repeat):
        if isinstance(repeat.counter, ast.Forever):
            counter = itertools.count(start=1)
        else:
            counter = range(1, repeat.counter.eval(self) + 1)

        for _i in counter:
#           assert i < 1000
            try:
                self.execute_block(repeat.body)
            except self.Break:
                break

    def execute_RepeatWhile(self, repeat_while):
        while repeat_while.condition.eval(self):
            try:
                self.execute_block(repeat_while.body)
            except self.Break:
                break

    def execute_If(self, if_):
        # TODO this will need to be fixed using iteration over linked list (not
        # recursion) because of the markers
        if if_.condition.eval(self):
            self.execute_block(if_.body)
        elif if_.orelse is not None:
            if_.orelse.execute(self)

    def execute_Else(self, else_):
        self.execute_block(else_.body)

    def execute_CallFunction(self, call_function):
        try:
            function = self.functions[call_function.name]
        except KeyError:
            raise NameError(
                f'function {call_function.name} undefined '
                f'at line {call_function.line} block {call_function.cols[0]}')

        function.execute(engine)

    def execute_Assign(self, assign):
        self.variables[assign.name] = assign.expression.eval(self)

    def execute_Place(self, place):
        return self.place(place.direction)

    def execute_Step(self, step):
        return self.step(step.counter.eval(self))

    def execute_Activate(self, activate):
        return self.activate()

    def execute_Draw(self, draw):
        return self.draw()

    def execute_Jump(self, jump):
        return self.jump()

    def execute_PickUp(self, pick_up):
        return self.pick_up()

    def execute_TurnLeft(self, turn_left):
        return self.turn_left()

    def execute_TurnRight(self, turn_right):
        return self.turn_right()

    def execute_Break(self, break_):
        raise self.Break()

    def execute_BreakFunction(self, break_function):
        raise self.BreakFunction()


    #
    # Abstract methods to be defined by game implementation
    # (e.g., pygame or physical robot)
    #

    @abstractmethod
    def step(self, counter: int):
        raise NotImplementedError()

    @abstractmethod
    def turn_left(self):
        raise NotImplementedError()

    @abstractmethod
    def turn_right(self):
        raise NotImplementedError()

    @abstractmethod
    def jump(self):
        raise NotImplementedError()

    @abstractmethod
    def pick_up(self):
        raise NotImplementedError()

    @abstractmethod
    def place(self, direction: ast.Direction):
        raise NotImplementedError()

    @abstractmethod
    def activate(self):
        raise NotImplementedError()

    @abstractmethod
    def draw(self):
        raise NotImplementedError()

    @abstractmethod
    def detect(self, direction: ast.Direction, obj: ast.Object):
        raise NotImplementedError()

    def noop(self):
        pass


class AsyncEngine(ExecEngine):
    async def eval_Number(self, number):
        return number.value

    async def eval_Variable(self, variable):
        try:
            return self.variables[variable.name]
        except KeyError:
            raise NameError(
                f'variable {self.varname} referenced before assignment '
                f'at line {self.line} block {self.cols[0]}')

    async def eval_BinOp(self, binop):
        return binop.op(await binop.left.eval(engine), await binop.right.eval(engine))

    async def eval_Condition(self, condition):
        logger.debug(f'{type(self).__name__}.eval_Condition()')
        return await self.detect(condition.direction, condition.object)


    async def execute_block(self, block):
        logger.debug(f'{type(self).__name__}.execute_block()')
        # TODO bracket
        for stmt in block:
            await stmt.execute(self)

    async def execute_Programme(self, programme):
        logger.debug(f'{type(self).__name__}.execute_Programme()')
        await self.execute_block(programme.body)

    async def execute_Repeat(self, repeat):
        logger.debug(f'{type(self).__name__}.execute_Repeat()')
        if isinstance(repeat.counter, ast.Forever):
            counter = itertools.count(start=1)
        else:
            counter = range(1, await repeat.counter.eval(self) + 1)

        for _i in counter:
#           assert i < 1000
            try:
                await self.execute_block(repeat.body)
            except self.Break:
                break

    async def execute_RepeatWhile(self, repeat_while):
        logger.debug(f'{type(self).__name__}.execute_RepeatWhile()')
        while await repeat_while.condition.eval(self):
            try:
                await self.execute_block(repeat_while.body)
            except self.Break:
                break

    async def execute_If(self, if_):
        # TODO this will need to be fixed using iteration over linked list (not
        # recursion) because of the markers
        if await if_.condition.eval(self):
            await self.execute_block(if_.body)
        elif if_.orelse is not None:
            await if_.orelse.execute(self)

    async def execute_Else(self, else_):
        await self.execute_block(else_.body)

    async def execute_CallFunction(self, call_function):
        try:
            function = self.functions[call_function.name]
        except KeyError:
            raise NameError(
                f'function {call_function.name} undefined '
                f'at line {call_function.line} block {call_function.cols[0]}')

        await function.execute(engine)

    async def execute_Assign(self, assign):
        self.variables[assign.name] = await assign.expression.eval(self)

    async def execute_Place(self, place):
        return await self.place(place.direction)

    async def execute_Step(self, step):
        return await self.step(await step.counter.eval(self))

    async def execute_Activate(self, activate):
        return await self.activate()

    async def execute_Draw(self, draw):
        return await self.draw()

    async def execute_Jump(self, jump):
        return await self.jump()

    async def execute_PickUp(self, pick_up):
        return await self.pick_up()

    async def execute_TurnLeft(self, turn_left):
        return await self.turn_left()

    async def execute_TurnRight(self, turn_right):
        return await self.turn_right()

    async def execute_Break(self, break_):
        raise self.Break()

    async def execute_BreakFunction(self, break_function):
        raise self.BreakFunction()


    #
    # Abstract methods to be defined by game implementation
    # (e.g., pygame or physical robot)
    #

    @abstractmethod
    async def step(self, counter: int):
        raise NotImplementedError()

    @abstractmethod
    async def turn_left(self):
        raise NotImplementedError()

    @abstractmethod
    async def turn_right(self):
        raise NotImplementedError()

    @abstractmethod
    async def jump(self):
        raise NotImplementedError()

    @abstractmethod
    async def pick_up(self):
        raise NotImplementedError()

    @abstractmethod
    async def place(self, direction: ast.Direction):
        raise NotImplementedError()

    @abstractmethod
    async def activate(self):
        raise NotImplementedError()

    @abstractmethod
    async def draw(self):
        raise NotImplementedError()

    @abstractmethod
    async def detect(self, direction: ast.Direction, obj: ast.Object):
        raise NotImplementedError()

    def noop(self):
        pass


class RemoteControlEngine(AsyncEngine):
    def __init__(self, *args, bot, **kwds):
        super().__init__(*args, **kwds)
        self.bot = bot

    async def step(self, counter: int):
        logger.debug(f'{type(self).__name__}.step({counter=})')
        r = range(abs(counter))
        logger.debug(f'  {r=}')
        for i in r:
            command = (bluetooth.Command.FORWARD if counter > 0
                else bluetooth.Command.BACKWARD)
            logger.debug(f'  {command=}')
            await asyncio.sleep(0.1)
            await self.bot.rpc(command)
        logger.debug(f'{type(self).__name__}.step() return')

    async def turn_left(self):
        await self.bot.rpc(bluetooth.Command.TURN_LEFT)

    async def turn_right(self):
        await self.bot.rpc(bluetooth.Command.TURN_RIGHT)

    async def jump(self):
        raise NotImplementedError()

    async def pick_up(self):
        raise NotImplementedError()

    async def place(self, direction: ast.Direction):
        raise NotImplementedError()

    async def activate(self):
        raise NotImplementedError()

    async def draw(self):
        raise NotImplementedError()

    async def detect(self, direction: ast.Direction, obj: ast.Object):
        logger.debug(f'{type(self).__name__}.detect({direction=}, {obj=})')
        if direction is not ast.Direction.IN_FRONT:
            raise NotImplementedError()

        if obj not in (ast.Object.OBSTACLE, ast.Object.EMPTY):
            raise NotImplementedError()

        sonar_range = await self.bot.sonar()
        logger.debug(f'{type(self).__name__}.detect {sonar_range=}')
        result = (sonar_range < 10) ^ (obj is ast.Object.EMPTY)
        logger.debug(f'{type(self).__name__}.detect -> {result}')
        return result


# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import pprint

import click
import click_default_group
import cv2
import pygame
import pygame.display

from loguru import logger

from . import (
    detector as _detector,
    game_bt,
    bluetooth,
)

@click.group(cls=click_default_group.DefaultGroup, default='game',
    default_if_no_args=True)
def cli():
    pass

@cli.command('game')
@click.option('--mac', required=True)
def game(mac):
    logger.debug('pygame.init()')
    pygame.init()
    pygame.display.set_caption('tuxgo')
    screen = pygame.display.set_mode(flags=pygame.FULLSCREEN)
    bot = bluetooth.BluetoothBot(mac)
    game = game_bt.Game(screen, bot)
    game.run()


@cli.command('debug-image')
@click.argument('inpath', type=click.Path(file_okay=True, dir_okay=False))
@click.argument('outpath', type=click.Path(file_okay=True, dir_okay=False))
def debug_image(inpath, outpath):
    detector = _detector.Detector()
    frame = cv2.imread(inpath)
    #frame = cv2.blur(frame, (5, 5))
    detections = list(detector.detect_markers(frame))
    analyser = _detector.Analyser(detections)
    try:
        programme = list(analyser.get_programme(debug_frame=frame))
        pprint.pprint(programme)
    finally:
        cv2.imwrite(outpath, frame)

@cli.command('debug-video')
def debug_video():
    vc = cv2.VideoCapture(0)
    detector = _detector.Detector()

    try:
        while True:
            ret, frame = vc.read()
            if not ret:
                click.echo('frame dropped', err=True)
                continue

            flipped = cv2.flip(frame, 1)

            for detection in detector.detect_markers(frame):
                detection.debug_draw_flipped(flipped)

            cv2.imshow('tuxgo-debug', flipped)

            key = cv2.waitKey(1)
            if key & 0xff == ord('q'):
                break
            elif key & 0xff == ord('f'):
                cv2.imwrite('screenshot.png', flipped)

    finally:
        vc.release()
        cv2.destroyAllWindows()

if __name__ == '__main__':
    cli()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import collections
import errno
import os
import pathlib
import time

class CPULoad:
    # see also: proc(5)

    def __init__(self):
        self._last = collections.defaultdict(lambda: (0, 0))
        self.usage = {}

    def update(self):
        with open('/proc/stat') as file:
            for line in file:
                if not line.startswith('cpu'):
                    continue
                cpu = line[3]
                cpu = None if cpu == ' ' else int(cpu)
                if cpu is None:
                    continue
                data = list(map(int, line.strip().split()[1:]))
                curr_sum, curr_idle = sum(data), data[3]
                last_sum, last_idle = self._last[cpu]
                try:
                    self.usage[cpu] = 100 * (
                        1 - (curr_idle - last_idle) / (curr_sum - last_sum))
                except ZeroDivisionError:
                    self.usage[cpu] = None
                self._last[cpu] = curr_sum, curr_idle

    def format_cpu(self, cpu):
        value = self.usage[cpu]
        value = '---.-' if value is None else f'{value:5.1f}'
        return f'CPU{cpu}: {value} %'

    def format_all(self):
        return ' '.join(self.format_cpu(cpu) for cpu in sorted(self.usage))


class _PWMProp:
    def __set_name__(self, owner, name):
        self.name = name
    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        data = (instance.path / self.name).read_text().strip()
        return int(data) if data.isdigit() else data
    def __set__(self, instance, value):
        (instance.path / self.name).write_text(str(value))

PWM_OPEN_RETRY_COUNT = 5

class PWM:
    def __init__(self, chip='pwmchip0', channel=0):
        self.path_chip = pathlib.Path('/sys/class/pwm') / chip
        try:
            (self.path_chip / 'export').write_text(str(channel))
        except OSError as e:
            if e.errno != errno.EBUSY:
                raise
        self.path = self.path_chip / f'pwm{channel}'

        # wait for udev to settle
        for i in range(PWM_OPEN_RETRY_COUNT):
            if os.access(self.path / 'enable', os.W_OK):
                break
            time.sleep(.1)

    duty_cycle = _PWMProp()
    enable = _PWMProp()
    period = _PWMProp()
    polarity = _PWMProp()

    @property
    def percent(self):
        return 100 * self.duty_cycle / self.period


def get_rfkill_state(name):
    for path in pathlib.Path('/sys/class/rfkill').glob('*'):
        if (path / 'name').read_text().strip() == name:
            return not int((path / 'state').read_text().strip())



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import datetime
import enum
import sys

import click
import cv2
import funcparserlib.parser
import libcamera
import numpy as np
import picamera2
import pygame
import pygame.camera
import pygame.display
import pygame.event
import pygame.surfarray
import pygame.time

from abc import abstractmethod
from loguru import logger

from . import (
    assets,
    badges,
    bluetooth,
    detector as _detector,
    parser,
    utils,
)

from .theme import Colour



PYGAME_EVENT_UPDATE_CPU_LOAD = pygame.event.custom_type()


class Widgets:
    def __init__(self, surface):
        self.surface = surface
        self.surface_width, self.surface_height = self.surface.get_size()
        self.font = pygame.font.Font(assets / 'fonts/VCROCDFaux.ttf', 30)

    def label(self, text, /, *,
        fg=Colour.WHITE, bg=Colour.BLACK,
        top=None, bottom=None, left=None, right=None,
    ):
        assert (top is not None) ^ (bottom is not None)
        assert (left is not None) ^ (right is not None)

        fontsurf = self.font.render(text, True, fg, bg)
        width, height = fontsurf.get_size()

        x = left if left is not None else self.surface_width - width - right
        y = (top * height if top is not None
            else self.surface_height - (bottom + 1) * height)
        self.surface.blit(fontsurf, (x, y))

    def key(self, key, descr, /, *,
        top=None, bottom=None, left=None, right=None,
    ):
        assert (top is not None) ^ (bottom is not None)
        assert (left is not None) ^ (right is not None)

        keysurf = self.font.render(key, True, Colour.WHITE, Colour.BLUE)
        descrsurf = self.font.render(' ' + descr, True, Colour.BLACK, Colour.WHITE)
        width, height = descrsurf.get_size()

        x = left if left is not None else self.surface_width - width - right
        y = (top * height if top is not None
            else self.surface_height - (bottom + 1) * height)

        self.surface.blit(descrsurf, (x, y))
        self.surface.blit(keysurf, (x - keysurf.get_width(), y))


class Component:
    def __init__(self, game):
        self.game = game

    def add_widgets(self):
        pass

    def handle_event(self, event):
        pass


class RobotComponent(Component):
    def add_widgets(self):
        if self.game.bot.connected:
            self.game.widgets.label(f'CONNECTED',
                bottom=2, left=0, bg=Colour.GREEN2)
            self.game.widgets.key('WSAD', 'Move robot',
                bottom=1, left=100)
            self.game.widgets.key('E', 'Sonar',
                bottom=1, left=350)

            if self.game.bot.last_range is not None:
                self.game.widgets.label(f'RANGE: {self.game.bot.last_range:3d} cm',
                    bottom=2, left=350)
        elif self.game.bot.loop is not None:
            self.game.widgets.label(f'CONNECTING',
                bottom=2, left=0)
        else:
            self.game.widgets.label(f'DISCONNECTED',
                bottom=2, left=0, bg=Colour.RED)
            self.game.widgets.key('R', 'Reconnect',
                bottom=1, left=100)

    def handle_event(self, event):
        if event.type != pygame.KEYDOWN:
            return

        if self.game.bot.connected:
            if event.key == pygame.K_w:
                self.game.bot.forward_nowait()
                return True
            if event.key == pygame.K_s:
                self.game.bot.backward_nowait()
                return True
            if event.key == pygame.K_a:
                self.game.bot.turn_left_nowait()
                return True
            if event.key == pygame.K_d:
                self.game.bot.turn_right_nowait()
                return True
            if event.key == pygame.K_e:
                self.game.bot.sonar_nowait()
                return True

        elif self.game.bot.loop is None:
            if event.key == pygame.K_r:
                self.game.bot.connect()
                return True

class BadgesComponent(Component):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.enabled = not all(badge is None for badge in self.game.badges)

    def add_widgets(self):
        if self.enabled:
            self.game.widgets.key('B', 'Badges', bottom=1, left=850)

    def handle_event(self, event):
        if (self.enabled
        and event.type == pygame.KEYDOWN
        and event.key == pygame.K_b):
            self.game.mode = self.game.mode_badges
            return True

class ClockComponent(Component):
    def add_widgets(self):
        self.game.widgets.label(f'{datetime.datetime.now():%H:%M}',
            top=0, right=0)

class CPUComponent(Component):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.cpuload = utils.CPULoad()
        self.cpuload.update()
        pygame.time.set_timer(PYGAME_EVENT_UPDATE_CPU_LOAD, 1000)

    def add_widgets(self):
        self.game.widgets.label(self.cpuload.format_all(), top=0, left=0)

    def handle_event(self, event):
        if event.type == PYGAME_EVENT_UPDATE_CPU_LOAD:
            self.cpuload.update()
            return True

class PWMComponent(Component):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.pwm = utils.PWM()
        self.pwm.period = 1000000 # ns -> 1 kHz
        self.pwm.duty_cycle = 0
        self.pwm.enable = 1

    def add_widgets(self):
        self.game.widgets.key('[]', 'Lamp ctl.', bottom=0, left=600)

    def handle_event(self, event):
        if event.type != pygame.KEYDOWN:
            return

        if event.key == pygame.K_LEFTBRACKET:
            self.pwm.duty_cycle = max(0, self.pwm.duty_cycle - 100000)
            return True
        if event.key == pygame.K_RIGHTBRACKET:
            self.pwm.duty_cycle = min(1000000, self.pwm.duty_cycle + 100000)
            return True

class RFKillComponent(Component):
    def add_widgets(self):
        if not utils.get_rfkill_state('phy0'):
            self.game.widgets.label('WIFI NOT RFKILL\'D', bottom=3, left=0, bg=Colour.RED)

class Game:
    def __init__(self, screen, bot):
        self.screen = screen
        self.screen_width, self.screen_height = self.screen.get_size()
        self.font_vcr_60 = pygame.font.Font(assets / 'fonts/VCROCDFaux.ttf', 60)
        self.widgets = Widgets(self.screen)
        self.clock = pygame.time.Clock()

        self.mode = self.mode_normal

        self.bot = bot

        self.camera = None
        self.detector = _detector.Detector()
        self.frame = None
        self.metadata = None

        self.badges = list(badges.Badge.load_from_assets())
        self.current_badge = None

        self.comp_clock = ClockComponent(self)
        self.comp_robot = RobotComponent(self)
        self.comp_badges = BadgesComponent(self)
        self.comp_cpu = CPUComponent(self)
        self.comp_pwm = PWMComponent(self)
        self.comp_rfkill = RFKillComponent(self)


    def run(self):
        self.init_camera()
        self.camera.start()
        try:
            while True:
                self.mode()
                pygame.display.flip()
                self.clock.tick(50)

        finally:
            self.comp_pwm.pwm.enable = 0
            try:
                self.camera.stop()
            except SystemError: # raised on double stop
                pass


    #
    # CAMERA FUNCTIONS
    #

    def init_camera(self):
        self.camera = picamera2.Picamera2()
        self.camera.configure(self.camera.create_preview_configuration(
            main={
                'format': 'BGR888',
                'size': (int(self.screen_height * 1.3), int(self.screen_width * 1.3)),
            },
            lores=None,
            raw=None,
            display=None,
            encode=None,
            controls={
                'ExposureValue': 1.1,
            },
        ))

    def capture_frame(self):
        request = self.camera.capture_request()
        self.frame = np.rot90(request.make_array('main')).copy()
        self.metadata = picamera2.Metadata(request.get_metadata())
        request.release()


    #
    # COMPONENTS AND EVENT HANDLING
    #

    def handle_exit(self, event):
        if (event.type == pygame.QUIT
        or (event.type == pygame.KEYDOWN and event.key == pygame.K_q)):
            if self.bot.loop is not None:
                self.bot.disconnect()
            sys.exit()

    def handle_event(self, components, event):
        self.handle_exit(event)

        # CPULoad is a bit special, we need to always catch the timer event,
        # no matter which mode we're in
        if self.comp_cpu.handle_event(event):
            return

        for component in components:
            if component.handle_event(event):
                break

    def add_widgets(self, components):
        for component in components:
            component.add_widgets()

    #
    # SCREEN MODES
    #

    def mode_normal(self):
        components = (
            self.comp_clock,
            self.comp_robot,
            self.comp_badges,
            self.comp_cpu,
            self.comp_pwm,
            self.comp_rfkill,
        )

        self.capture_frame()
        display_frame = self.frame.copy()

        detections = self.detector.detect_markers(self.frame)
        analyser = _detector.Analyser(detections)
        for detection in detections:
            detection.debug_draw(display_frame)
        try:
            programme = list(analyser.get_programme(debug_frame=display_frame))
        except _detector.BoardError:
            error = True
        else:
            error = False

#           surface = pygame.surfarray.make_surface(frame.T)

        # XXX WTF this math, TODO get understanding
        surface = pygame.surfarray.make_surface(np.flipud(np.rot90(display_frame)))
        surface = pygame.transform.scale(surface, self.screen.get_size())
        frame_width, frame_height = surface.get_size()
#       logger.debug(f'{surface.get_size()=}')

        self.screen.blit(surface, (
            (self.screen_width  - frame_width)  / 2, 
            (self.screen_height - frame_height) / 2))

        self.widgets.label(f'EXP: {self.metadata.ExposureTime:5} us',
            top=1, left=0)
        self.widgets.label(f'GAIN:'
            f' A{self.metadata.AnalogueGain:.1f}'
            f' D{self.metadata.DigitalGain:.1f}'
            f' R{self.metadata.ColourGains[0]:.1f}'
            f' B{self.metadata.ColourGains[1]:.1f}',
            top=1, left=240)
        self.widgets.label(f'BLUR: {self.detector.blur} px',
            top=1, left=650)
        self.widgets.label(f'FPS: {self.clock.get_fps():4.1f} Hz',
            top=1, right=0)

        self.widgets.label(f'SHUTTER: 1/{1000000/self.metadata.ExposureTime:3.0f} s',
            top=2, left=0)
        self.widgets.label(f'ILLUM: {self.metadata.Lux:4.0f} lux',
            top=2, left=300)
        self.widgets.label(f'PWM: {self.comp_pwm.pwm.percent:3.0f} %',
            top=2, left=600)

        self.widgets.label(f'TEMP: {self.metadata.SensorTemperature:2.0f} °C',
            top=2, right=0)



        if error:
            self.widgets.label('ERROR', bottom=2, right=0, bg=Colour.RED)
        else:
            self.widgets.label('OK', bottom=2, right=0)

        self.widgets.key('SPACE', 'Snapshot', bottom=0, left=100)

#       self.widgets.key('C', 'Cam. sett.', bottom=1, left=600)
#       self.widgets.key('I', 'Load file', bottom=0, left=600)
        self.widgets.key('+-', 'Blur ctl.', bottom=1, left=600)

        self.widgets.key('Q', 'Quit', bottom=0, left=850)

        self.add_widgets(components)
#       for i, item in enumerate(self.metadata.__dict__.items()):
#           k, v = item
#           self.widgets.label(f'{k}: {v}', left=0, top=i + 3)


        for event in pygame.event.get():
            self.handle_event(components, event)

            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_SPACE:
                    self.mode = self.mode_snapshot

                if event.key == pygame.K_MINUS:
                    self.detector.blur = max(1, self.detector.blur - 1)
                if event.key in (pygame.K_PLUS, pygame.K_EQUALS):
                    self.detector.blur += 1


    def mode_snapshot(self):
        components = (
            self.comp_clock,
            self.comp_robot,
            self.comp_badges,
            self.comp_pwm,
            self.comp_rfkill,
        )

        display_frame = self.frame.copy()
        detections = self.detector.detect_markers(self.frame)
        analyser = _detector.Analyser(detections)
        for detection in detections:
            detection.debug_draw(display_frame)
        try:
            programme = list(analyser.get_programme(debug_frame=display_frame))
            parsed = parser.parse(programme)

        except (_detector.BoardError, funcparserlib.parser.NoParseError):
            error = True
        else:
            error = False

#           surface = pygame.surfarray.make_surface(frame.T)

        # XXX WTF this math, TODO get understanding
        surface = pygame.surfarray.make_surface(np.flipud(np.rot90(display_frame)))
        surface = pygame.transform.scale(surface, self.screen.get_size())
        frame_width, frame_height = surface.get_size()
#       logger.debug(f'{surface.get_size()=}')

        self.screen.blit(surface, (
            (self.screen_width  - frame_width)  / 2, 
            (self.screen_height - frame_height) / 2))

        self.add_widgets(components)
        self.widgets.key('SPACE', 'Snapshot', bottom=0, left=100)
        self.widgets.key('Q', 'Quit', bottom=0, left=850)

        if error:
            self.widgets.label('ERROR', bottom=2, right=0, bg=Colour.RED)
        else:
            for i, line in enumerate(programme, 1):
                line = ' '.join(token.value.text for token in line)
                self.widgets.label(f'{i:2d} {line}', top=i, left=0)
            self.widgets.label('OK', bottom=2, right=0)

            if self.bot.connected:
                if self.bot.current_task is not None:
                    self.widgets.key('X', 'Stop', bottom=0, left=350)
                elif not error:
                    self.widgets.key('R', 'Run', bottom=0, left=350)

        for event in pygame.event.get():
            self.handle_event(components, event)

            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    self.mode = self.mode_normal

                if event.key == pygame.K_SPACE:
                    self.capture_frame()

                if self.bot.connected:
                    if self.bot.current_task is not None:
                        if event.key == pygame.K_x:
                           self.bot.stop_programme()
                    elif not error:
                        if event.key == pygame.K_r:
                           self.bot.execute_programme(parsed)

    def mode_badges(self):
        self.screen.fill((0, 0, 0))

        image_size = 256
        assert self.screen_height == 1280

        padding = (self.screen_width - 2 * image_size) / 3

        for i, badge in enumerate(self.badges):
            if badge is None:
                continue

            x = padding + (image_size + padding) * (i // 5)
            y = image_size * (i % 5)

            image = pygame.transform.scale(
                badge.image_menu or badge.image, (image_size, image_size))

            label = self.font_vcr_60.render(
                str((i + 1) % 10), True, (0, 0, 0), (255, 255, 255))

            self.screen.blit(image, (x, y))
            self.screen.blit(label, (x - label.get_width() - 16, y))

        for event in pygame.event.get():
            self.handle_event((), event)

            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    self.mode = self.mode_normal
                    continue

                for i in range(min(len(self.badges), 10)):
                    if event.key == getattr(pygame, f'K_{(i + 1) % 10}'):
                        if self.badges[i] is not None:
                            self.mode = self.mode_badge
                            self.current_badge = self.badges[i]

    def mode_badge(self):
        assert self.current_badge

        padding = 48

        self.screen.fill((0, 0, 0))

        qrcode_size = (self.screen_height
            - 3 * padding - self.current_badge.image.get_height())
        qrcode = pygame.transform.scale(
            self.current_badge.qrcode, (qrcode_size, qrcode_size))

        self.screen.blit(qrcode,
            ((self.screen_width - qrcode_size) / 2, padding))
        self.screen.blit(self.current_badge.image, (
            (self.screen_width - self.current_badge.image.get_width()) / 2,
            (self.screen_height - self.current_badge.image.get_height() - padding)
        ))

        for event in pygame.event.get():
            self.handle_event((), event)

            if event.type == pygame.KEYDOWN:
                if event.key in (pygame.K_ESCAPE, pygame.K_b):
                    self.mode = self.mode_badges
                    self.current_badge = None
                    continue

                for i in range(min(len(self.badges), 10)):
                    if event.key == getattr(pygame, f'K_{(i + 1) % 10}'):
                        if self.badges[i] is not None:
                            self.current_badge = self.badges[i]

    def mode_camera_settings(self):
        raise NotImplementedError()

    def mode_load_file(self):
        raise NotImplementedError()


# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import asyncio
import enum
import io
import math
import pathlib
import sys

import click
import cv2
import numpy as np
import pygame
import matplotlib.path
import matplotlib.transforms

from . import (
    assets,
    ast,
    blocks,
    detector,
    engine,
    http_server,
    parser,
)

def load_svg(path, size):
    # Robot assets have width="100%" height="100%" attributes on svg, and pygame
    # loads them as 1x1x32 surfaces, which is useless. This is adapted
    # https://stackoverflow.com/a/69509545.
    path = pathlib.Path(path)
    width, height = size
    with open(path, 'rb') as file:
        data = file.read()
    i = data.index(b'">') + 1 # this will find the end of <svg> opening tag

    buffer = io.BytesIO()
    buffer.write(data[:i])
    buffer.write(f' width="{width}" height="{height}"'.encode('ascii'))
    buffer.write(data[i:])
    buffer.seek(0)

    return pygame.image.load(buffer, path.name)

TILE_WIDTH = 128

class LevelField(enum.Enum):
    ABYSS = ' '
    FLOOR = '.' # aka EMPTY
    WALL = '#'
    ELEVATED = ':'
    WATER = '~'
    LIGHT_OFF = '?'
    LIGHT_ON = '!'

_TILESET = {
#   LevelField.FLOOR:
#       load_svg(assets / 'tileset/401.svg', (TILE_WIDTH, TILE_WIDTH * 3//4)),
    LevelField.FLOOR:
        load_svg(assets / 'tileset/001.svg', (TILE_WIDTH, TILE_WIDTH * 3//4)),
    LevelField.WALL:
        load_svg(assets / 'tileset/472.svg', (TILE_WIDTH, TILE_WIDTH * 11//4)),
    LevelField.ELEVATED:
        load_svg(assets / 'tileset/477.svg', (TILE_WIDTH, TILE_WIDTH * 11//4)),
    LevelField.LIGHT_OFF:
        load_svg(assets / 'tileset/051.svg', (TILE_WIDTH, TILE_WIDTH * 11//4)),
    LevelField.LIGHT_ON:
        load_svg(assets / 'tileset/052.svg', (TILE_WIDTH, TILE_WIDTH * 11//4)),
}


# 37c3 colours
BLACK =     (  0,   0,   0) #000000
WHITE =     (255, 255, 255) #ffffff
GREY1 =     (217, 217, 217) #d9d9d9
GREY2 =     (170, 170, 170) #aaaaaa
GREY3 =     (122, 122, 122) #7a7a7a
GREY4 =     ( 32,  32,  32) #202020
BLUE =      ( 45,  66, 255) #2d42ff
BLUE2 =     ( 11,  21, 117) #0b1575
RED =       (222,  64,  64) #de4040
RED2 =      ( 86,  16,  16) #561010
GREEN =     (121, 255,  94) #79ff5e
GREEN2 =    ( 43, 141,  24) #2b8d18
CYAN =      ( 41, 255, 255) #29ffff
CYAN2 =     (  0, 107, 107) #006b6b
MAGENTA =   (222,  55, 255) #de37ff
MAGENTA2 =  (102,   0, 122) #66007a
YELLOW =    (246, 246, 117) #f6f675
YELLOW2 =   (117, 177,   1) #757501


class Level:
    def __init__(self, tiles):
        self.tiles = tiles

        # diagonal or "major" diagonal is from 0,0 to +inf,+inf 
        # minor diagonal is from 0,+inf to +inf,0
        self.min_major_diag = self.min_minor_diag = float('+inf')
        self.max_major_diag = self.max_minor_diag = float('-inf')

        for y in range(len(tiles)):
            for x in range(len(tiles[y])):
                if self.tiles[y][x] == LevelField.ABYSS:
                    continue
                major_diag, minor_diag = x + y, y - x
                self.min_major_diag = min(self.min_major_diag, major_diag)
                self.max_major_diag = max(self.max_major_diag, major_diag)
                self.min_minor_diag = min(self.min_minor_diag, minor_diag)
                self.max_minor_diag = max(self.max_minor_diag, minor_diag)

        assert not any(math.isinf(i) for i in (
            self.min_major_diag,
            self.max_major_diag,
            self.min_minor_diag,
            self.max_minor_diag,
        ))

    @classmethod
    def from_file(cls, file):
        tiles = []
        hero = None
        for y, line in enumerate(file):
            line = line.rstrip('\n')
            if not line:
                break
            tileline = []
            for x, c in enumerate(line):
                if c == '@':
                    if hero is not None:
                        raise ValueError('more than one hero')
                    hero = (x, y)
                    c = '.'
                tileline.append(LevelField(c))
            tiles.append(tileline)
        return cls(tiles), hero


class Button:
    def __init__(self, surface, font, x, y, width, height, text):
        self.surface = surface
        self.font = font
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.text = text

    def draw(self):
        radius = self.height / 2
        pygame.draw.circle(self.surface, WHITE, (self.x + radius, self.y + radius), radius)
        pygame.draw.rect(self.surface, WHITE, pygame.Rect((self.x + radius, self.y), (self.width - radius, self.height)))
        font_surf = self.font.render(self.text, True, (0, 0, 0), WHITE)
        self.surface.blit(font_surf, (self.x + radius * 0.7, self.y + (self.height - font_surf.get_height()) / 2))

class ButtonSprite(pygame.sprite.Sprite):
    def __init__(self, x, y, width, height, text, font):
        super().__init__()

        self.image = pygame.Surface([width, height])
        self.image.fill((0, 0, 0))
        self.rect = self.image.get_rect()
        self.rect.x, self.rect.y = x, y

        radius = height / 2
        pygame.draw.circle(self.image, WHITE, (radius, radius), radius)
        pygame.draw.rect(self.image, WHITE, pygame.Rect((radius, 0), (width - radius, height)))
        text_surf = font.render(text, True, (0, 0, 0), WHITE)
        self.image.blit(text_surf, (radius * 0.7, (height - text_surf.get_height()) / 2))

class ButtonBar(pygame.sprite.AbstractGroup):
    width = 300
    height = 100

    def __init__(self, screen, n=7):
        super().__init__()
        self._buttons = [None for i in range(n)]

        screen_rect = screen.get_rect()
        self.x = screen_rect.width - self.width
        self.padding = (screen_rect.height - n * self.height) / (n + 1)
        self.font = pygame.font.Font(assets / 'fonts/Rajdhani-Bold.otf', 60)

    def __getitem__(self, key):
        return self._buttons[key]

    def set_button(self, index, text, _callback):
        if self._buttons[index] is not None:
            self._buttons[index].kill()
        self.add(ButtonSprite(self.x, self.y(index), self.width, self.height, text, self.font), index=index)

    def remove_button(self, i):
        self._buttons[i].sprite.kill()

    def y(self, index):
        index %= len(self._buttons) # can be negative!
        return self.padding * (index + 1) + self.height * index

    def add(self, sprite, *, index):
        if self._buttons[index] is not None:
            self.remove(self._buttons[index])
        super().add(sprite)
        self._buttons[index] = sprite

    def remove_internal(self, sprite):
        self._buttons[self._buttons.index(sprite)] = None
        super().remove_internal(self, sprite)

class Tile(pygame.sprite.Sprite):
    def __init__(self, *args, image, midbottom, **kwds):
        super().__init__(*args, **kwds)
        self.image = image
        self.rect = image.get_rect()
        self.rect.midbottom = midbottom

class Robot(pygame.sprite.Sprite):
    def __init__(self, *args, transform_level, images, pos, rotation, **kwds):
        super().__init__(*args, **kwds)
        self.images = images
        self.transform = matplotlib.transforms.CompositeAffine2D(
            transform_level,
            matplotlib.transforms.Affine2D().translate(0, -36)
        )
        self._rotation = None
        self.set_rotation(rotation)
        self.rect = self.image.get_rect()
        self._pos = None
        self.move(pos)

    @property
    def pos(self):
        return self._pos

    def move(self, pos):
        self._pos = tuple(pos)
        self.rect.midbottom = self.transform.transform(self._pos)

        layer = pos_to_layer(*pos)
        for group in self.groups():
            try:
                change_layer = group.change_layer
            except AttributeError:
                pass
            else:
                change_layer(self, layer)

    @pos.setter
    def pos(self, pos):
        return self.move(pos)

    def set_rotation(self, rotation):
        self._rotation = rotation % len(self.images)
        self.image = self.images[self._rotation]

    def get_rotation(self):
        return self._rotation

    rotation = property(get_rotation, set_rotation)


def get_sprites_for_level(level, transform):
    group = pygame.sprite.LayeredUpdates()
    for y in range(len(level.tiles)):
        for x in range(len(level.tiles[y])):
            tile = level.tiles[y][x]
            if tile == LevelField.ABYSS:
                continue
            group.add(Tile(
                    image=_TILESET[level.tiles[y][x]],
                    midbottom=transform.transform((x, y))),
                layer=pos_to_layer(x, y))
    return group


# there's a stack of three transforms:
# 1. transform_isometric: base transform that skews in game (integral)
#    coordinates into isometric view
# 2. transform_level: translates the level into the centre of the screen; this
#    transform is applied to all the sprites
# 3. transform_robot: adds offset to the robot sprite (which is a bit special)
#    in relation to the board

# (0,0) is to the top
TRANSFORM_ISOMETRIC_ORIGIN_TOP = matplotlib.transforms.Affine2D.from_values(
     64, 32,
    -64, 32,
    0, 0,
)

# (0,0) is to the left
TRANSFORM_ISOMETRIC_ORIGIN_LEFT = matplotlib.transforms.Affine2D.from_values(
    64, -32,
    64,  32,
    0, 0,
)

def pos_to_layer_origin_top(x, y):
    return y + x

def pos_to_layer_origin_left(x, y):
    return y - x

pos_to_layer = pos_to_layer_origin_left

def get_level_transform_for_surface_origin_top(surface, level):
    half = surface.get_height() / 2
    tx = max(
        half + (level.min_minor_diag + level.max_minor_diag) / 2 * TILE_WIDTH / 2,
        (level.max_minor_diag + 1) * TILE_WIDTH / 2
    )
    ty = max(
        half - (level.min_major_diag + level.max_major_diag) / 2 * TILE_WIDTH / 4
            + TILE_WIDTH / 2,
        (-level.min_major_diag) * TILE_WIDTH / 2 + TILE_WIDTH * 3 / 4
    )

    return matplotlib.transforms.CompositeAffine2D(
        TRANSFORM_ISOMETRIC_ORIGIN_TOP,
        matplotlib.transforms.Affine2D().translate(tx, ty)
    )

def get_level_transform_for_surface_origin_left(surface, level):
    print(f'{level.min_major_diag=}')
    print(f'{level.max_major_diag=}')
    print(f'{level.min_minor_diag=}')
    print(f'{level.max_minor_diag=}')

    half = surface.get_height() / 2

    tx = max(
        half - (level.min_major_diag + level.max_major_diag) / 2 * TILE_WIDTH / 2,
        (level.min_major_diag + 1) * TILE_WIDTH / 2
    )
    ty = max(
        half - (level.min_minor_diag + level.max_minor_diag) / 2 * TILE_WIDTH / 4
            + TILE_WIDTH / 2,
        (level.min_minor_diag) * TILE_WIDTH / 2 + TILE_WIDTH * 3 / 4
    )

    return matplotlib.transforms.CompositeAffine2D(
        TRANSFORM_ISOMETRIC_ORIGIN_LEFT,
        matplotlib.transforms.Affine2D().translate(tx, ty)
    )

get_level_transform_for_surface = get_level_transform_for_surface_origin_left

class Game(engine.ExecEngine):
    def __init__(self, *args, screen, level, hero, hero_rotation, **kwds):
        self.screen = screen
        self.level = level
        self.transform_level = get_level_transform_for_surface(screen, level)
        self.robot = Robot(transform_level=self.transform_level, images=[
            load_svg(assets / 'robot/robot01.svg', (96, 96)),
            load_svg(assets / 'robot/robot02.svg', (96, 96)),
            load_svg(assets / 'robot/robot03.svg', (96, 96)),
            load_svg(assets / 'robot/robot04.svg', (96, 96)),
        ], pos=hero, rotation=hero_rotation)
        self.sprites = get_sprites_for_level(
            level, transform=self.transform_level)
        self.sprites.add(self.robot, layer=pos_to_layer(*self.robot.pos))
        self.running = False
        self.clock = pygame.time.Clock()

    def loop_once(self, tick=1, *, _really_once=True):
        self.process_events(once=_really_once)
        self.screen.fill((0, 0, 0))
        self.sprites.draw(self.screen)

        height = self.screen.get_height()
        half = height / 2
        pygame.draw.line(self.screen, (255, 0, 0),
            (0, half), (height, half))
        pygame.draw.line(self.screen, (255, 0, 0),
            (half, 0), (half, height))

        pygame.display.flip()
        if tick:
            self.clock.tick(tick)

    def loop(self):
        self.running = True

        while self.running:
            self.loop_once(tick=False)
            self.clock.tick(50)

    def process_events(self, once):
        for event in pygame.event.get():
            if (event.type == pygame.QUIT
            or event.type == pygame.KEYDOWN and event.key == pygame.K_q):
                if once:
                    sys.exit()
                else:
                    self.running = False


    def step(self, counter: int):
        pos = list(self.robot.pos)
        for i in range(abs(counter)):
            pos[self.robot.rotation % 2] += (int(math.copysign(1, counter))
                * ((-1) ** (self.robot.rotation // 2)))
            self.robot.pos = pos
            self.loop_once()

    def turn_left(self):
        self.robot.rotation -= 1
        self.loop_once()

    def turn_right(self):
        self.robot.rotation += 1
        self.loop_once()

    def jump(self):
        self.step(1)
#       self.loop_once()
#       raise NotImplementedError()

    def pick_up(self):
        raise NotImplementedError()

    def place(self, direction: ast.Direction):
        raise NotImplementedError()

    def activate(self):
        raise NotImplementedError()

    def draw(self):
        raise NotImplementedError()

    def detect(self, direction: ast.Direction, obj: ast.Object):
        raise NotImplementedError()

    def execute_Programme(self, programme):
        self.loop_once()
        super().execute_Programme(programme)
        self.loop()


w_przeręblu = [
    (blocks.Token.BEGIN,),
    (blocks.Token.REPEAT, blocks.Token.FOREVER),
        (blocks.Token.TURN_RIGHT,),
    (blocks.Token.END_BLOCK,),
    (blocks.Token.END,),
]

prog1_blocks = [
    (blocks.Token.BEGIN,),
    (blocks.Token.STEP, blocks.Token.DIGIT_1),
    (blocks.Token.TURN_RIGHT,),
    (blocks.Token.STEP, blocks.Token.DIGIT_1),
    (blocks.Token.TURN_LEFT,),
    (blocks.Token.STEP, blocks.Token.DIGIT_1),
    (blocks.Token.TURN_RIGHT,),
    (blocks.Token.STEP, blocks.Token.DIGIT_1),
    (blocks.Token.TURN_LEFT,),
    (blocks.Token.STEP, blocks.Token.DIGIT_1),
    (blocks.Token.TURN_RIGHT,),
    (blocks.Token.STEP, blocks.Token.DIGIT_1),
    (blocks.Token.TURN_LEFT,),
    (blocks.Token.END,),
]


def game_noninteractive():
    pygame.init()
    pygame.display.set_caption('tuxgo')
    screen = pygame.display.set_mode(flags=pygame.FULLSCREEN)

    with open(assets / 'levels/level-woju') as file:
#   with open(assets / 'levels/level-mem-sy') as file:
#   with open(assets / 'levels/level-0-0') as file:
        level, hero = Level.from_file(file)
    if hero is None:
        hero = (0, 0)

    vm = Game(
        screen=screen,
        level=level,
        hero=hero,
        hero_rotation=0,
    )

    parser.parse(w_przeręblu).execute(vm)
#   parser.parse(prog1_blocks).execute(vm)

    pygame.quit()

def game_interactive():
    det = detector.Detector()

    pygame.init()
    pygame.display.set_caption('tuxgo')
    screen = pygame.display.set_mode(flags=pygame.FULLSCREEN)

    with open(assets / 'levels/level-woju') as file:
#   with open(assets / 'levels/level-mem-sy') as file:
#   with open(assets / 'levels/level-0-0') as file:
        level, hero = Level.from_file(file)
    if hero is None:
        hero = (0, 0)

    vm = Game(
        screen=screen,
        level=level,
        hero=hero,
        hero_rotation=0,
    )

    vm.loop_once(tick=0)

    while True:
        image = http_server.get_image_from_http_server()
        image = np.asarray(bytearray(image), dtype=np.uint8)
        frame = cv2.imdecode(image, 0)
        analyser = detector.Analyser(det.detect_markers(frame))
        programme = parser.parse(analyser.get_programme())
        programme.execute(vm)

    pygame.quit()

def main():
    game_noninteractive()

if __name__ == '__main__':
    main()

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import tomllib

import cv2

import pygame
import pygame.image
import pygame.surfarray

from . import assets

class Badge:
    def __init__(self, qrcode, image, image_menu=None):
        self.qrcode = qrcode
        self.image = image
        self.image_menu = image_menu

    @classmethod
    def load_from_assets(cls):
        data = tomllib.load(open(assets / 'badges/badges.toml', 'rb'))
        url_template = data['url_template']

        for badge in data['badge']:
            if not badge:
                yield None
                continue

            url = url_template.format(token=badge.pop('token'))
            qrcode = pygame.surfarray.make_surface(
                cv2.QRCodeEncoder.create().encode(url))

            image_path = assets / 'badges' / badge.pop('image')
            image = pygame.image.load(image_path)

            try:
                image_menu_path = assets / 'badges' / badge.pop('image_menu')
            except KeyError:
                image_menu = None
            else:
                image_menu = pygame.image.load(image_menu_path)
            
            assert not badge
            yield cls(qrcode, image, image_menu)

if __name__ == '__main__':
    main()

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import sys
import time
import typing
import unicodedata

import cv2
import cv2.aruco
import matplotlib.path
import matplotlib.transforms
import numpy as np

from loguru import logger

from . import (
    ast,
    blocks,
)

class Affine2DWithAxisRotation(matplotlib.transforms.Affine2D):
    def rotate_to_axis(self, axis_unit_vector):
        vx, vy = axis_unit_vector
        rotate_mtx = np.array([
            [-vy,  vx,   0],
            [-vx, -vy,   0],
            [  0,   0,   1]], float)
        self._mtx = rotate_mtx @ self._mtx
        self.invalidate()
        return self

    def rotate_from_axis(self, axis_unit_vector):
        vx, vy = axis_unit_vector
        rotate_mtx = np.array([
            [-vy, -vx,   0],
            [ vx, -vy,   0],
            [  0,   0,   1]], float)
        self._mtx = rotate_mtx @ self._mtx
        self.invalidate()
        return self

def draw_label(img, text, org, fontFace, fontScale, color_bg, color_fg, thickness):
    if img is None:
        return None
    text = text.replace('Ł', 'L').replace('ł', 'l')
    text = ''.join(c for c in unicodedata.normalize('NFKD', text)
        if unicodedata.category(c)[0] != 'M')
    textsize, baseline = cv2.getTextSize(text, fontFace, fontScale, thickness)
    cv2.rectangle(
        img,
        (org[0] - 2, org[1] + baseline + 2),
        (org[0] + textsize[0] + 2, org[1] - textsize[1] - baseline - 2),
        color_bg,
        -1,
    )
    cv2.putText(img, text, org, fontFace, fontScale, color_fg, thickness)
    return img

def draw_path(img, path, color, thickness):
    if img is None:
        return None
    cv2.polylines(
        img, np.array(path.to_polygons()).astype(int), True, color, thickness)
    return img

class Detection(typing.NamedTuple):
    aruco_id: int
    corners: np.ndarray
    block: blocks.Block

    def _debug_draw(self, frame, corners, color_bg, color_fg):
        if frame is None:
            return

        cv2.polylines(frame, corners.reshape((1,4,2)).astype(int), True,
                color_bg, 2)

        top_left, *_ = corners
        cv2.rectangle(
            frame,
            top_left - [5, 5],
            top_left + [5, 5],
            color_bg,
            thickness=-1,
        )
        draw_label(
            frame,
            f'{self.aruco_id} ({self.block})',
            (top_left[0] + 10, top_left[1] + 5),
            cv2.FONT_HERSHEY_SIMPLEX,
            1,
            color_bg,
            color_fg,
            2,
        )

    def debug_draw(self, frame, color_bg=(0, 0, 0), color_fg=(255, 255, 255)):
        return self._debug_draw(frame, self.corners, color_bg, color_fg)

    def debug_draw_flipped(self, frame, color_bg=(0, 0, 0), color_fg=(255, 255, 255)):
        _, width, _ = frame.shape
        return self._debug_draw(frame, np.array([
            (width - x, y) for x, y in self.corners
        ]))

    def get_next_detect_areas(
        self, base_transform, axis_v, axis_h, *, debug_frame=None,
    ):
        top_left, _, _, bottom_left = self.corners
        marker_height = np.linalg.norm(top_left - bottom_left)
        corners_path = matplotlib.path.Path(self.corners)

        transform = (
              matplotlib.transforms.Affine2D()
                .translate(*-top_left)
            + base_transform
            + matplotlib.transforms.Affine2D()
                .translate(*top_left)
                .translate(*axis_v * marker_height * -0.6)
        )
        transform_v = (transform + matplotlib.transforms.Affine2D()
            .translate(*axis_v * marker_height * 2.2)
            .translate(*axis_h * marker_height * -1))

        transform_h = (transform + matplotlib.transforms.Affine2D()
            .translate(*axis_h * marker_height * 4.5)
        )

        next_area_v = transform_v.transform_path(corners_path)
        next_area_h = transform_h.transform_path(corners_path)

        return next_area_v, next_area_h


class Detector:
    def __init__(self, blur=1):
        self.blocks = blocks.BlockLoader()
        self.dictionary = cv2.aruco.getPredefinedDictionary(
            cv2.aruco.DICT_ARUCO_ORIGINAL)
        self.blur = blur
#       self.parameters = cv2.aruco.DetectorParameters_create()
#       self.parameters.detectInvertedMarker = True

    def detect_markers(self, frame):
        start_time = time.perf_counter()
        frame = 255 - frame
        if self.blur > 1:
            frame = cv2.blur(frame, (self.blur, self.blur))

        corners, ids, _rejected = cv2.aruco.detectMarkers(
            image=frame,
            dictionary=self.dictionary,
#           parameters=self.parameters,
        )

#       logger.debug(f'{rejected=}')

        if ids is not None:
            for c, aruco_id in zip(corners, ids.flatten()):
                try:
                    block = self.blocks.get_block_by_aruco_id(aruco_id)
                except LookupError:
                    block = None
                yield Detection(aruco_id, c.reshape((4, 2)).astype(int), block)

        perf_ms = (time.perf_counter() - start_time) * 1000
#       logger.debug(f'detect_markers() took {perf_ms:.3f} ms')

class BoardError(Exception):
    pass

class Analyser:
    def __init__(self, detections):
        self.detections = list(detections)

    def find_start_and_end(self):
        start = None
        end = None
        is_func = None

        for i in range(len(self.detections)-1, -1, -1):
            block = self.detections[i].block
            if block is None:
                continue
            token = block.token
#           logger.debug(f'{token=} {start=} {end=}')
            if token.type in (blocks.TokenType.BEGIN, blocks.TokenType.END):
                if token.type is blocks.TokenType.BEGIN:
                    if start is not None:
                        raise BoardError('double start tile')
                    start = self.detections[i]
                else:
                    if end is not None:
                        raise BoardError('double end tile')
                    end = self.detections[i]
                current_is_func = token in (blocks.Token.DEFINE_FUNCTION,
                        blocks.Token.END_FUNCTION)
                if is_func is not None and current_is_func != is_func:
                    raise BoardError('mismatched start and end')
                is_func = current_is_func
                del self.detections[i]

        if start is None:
            raise BoardError('no start tile')
        if end is None:
            raise BoardError('no start tile')

        return start, end

    def find_detection_in_area(self, area, *, debug_frame=None):
        """
        Find a single `Detection` inside *area* given by `matplotlib.path.Path`

        If there aren't any, return `None`. If there are more than one, raises
        `BoardError`.

        Any detections found are removed from self.detections.

        Debug drawing:

        - on found or `None`, draws cyan area and white-on-black detection
        - on error, draws red area and white-on-red detection
        """
        candidate = None
        found_multiple = False
        for i in range(len(self.detections)-1, -1, -1):
            i_top_left, *_ = self.detections[i].corners
            if area.contains_point(i_top_left):
                if found_multiple or candidate is not None:
                    found_multiple = True
                    self.detections.pop(i).debug_draw(debug_frame, (0, 0, 255))
                    continue

                candidate = self.detections.pop(i)

        if found_multiple:
            candidate.debug_draw(debug_frame, (0, 0, 255))
            draw_path(debug_frame, area, (0, 0, 255), 2)
            raise BoardError('more than one candidate in detection area')

        if candidate is not None:
            candidate.debug_draw(debug_frame)
        draw_path(debug_frame, area, (255, 255, 0), 2)
        return candidate

    def get_line(
        self, detection, area_h, base_transform, axis_v, axis_h, *,
        debug_frame=None
    ):
        while True:
            yield detection.block.token
            detection = self.find_detection_in_area(area_h,
                debug_frame=debug_frame)
            if detection is None:
                break
            _, area_h = detection.get_next_detect_areas(
                base_transform, axis_v, axis_h, debug_frame=debug_frame)

    def get_programme(self, *, debug_frame=None):
        start_time = time.perf_counter()

        # 1. find start and end tiles
        start, end = self.find_start_and_end()

        # 2. calculate main axis and unit vector parallel to axis
        start_top_left, *_ = start.corners
        end_top_left, *_ = end.corners

        if debug_frame is not None:
#           cv2.line(debug_frame, start_top_left, (0, 0), (255, 255, 0), 2)
            cv2.line(debug_frame, start_top_left, end_top_left, (255, 255, 0), 2)

        axis = end_top_left - start_top_left
        axis_v = axis / np.linalg.norm(axis)
        axis_h = np.array([axis_v[1], -axis_v[0]])

        # 3. calcucate transforms for rotating and scaling detection area
        # (will be applied to each marker outline around its top_left)
        # (will be used for both horizontal and vertical areas)
        base_transform = (Affine2DWithAxisRotation()
            .rotate_to_axis(axis_v)
            # 4 == difference between narrowest and widest tile + 1
            .scale(4, 1.2)
            .rotate_from_axis(axis_v)
        )

        # 4. yield all the lines
        try:
            current = start
            while True:
                detect_area_v, detect_area_h = current.get_next_detect_areas(
                    base_transform, axis_v, axis_h, debug_frame=debug_frame)
                yield tuple(self.get_line(current, detect_area_h,
                    base_transform, axis_v, axis_h, debug_frame=debug_frame))

                if detect_area_v.contains_point(end_top_left):
                    yield (end.block.token,)
                    if debug_frame is not None:
                        end.debug_draw(debug_frame)
                    break

                current = self.find_detection_in_area(detect_area_v,
                    debug_frame=debug_frame)

                if current is None:
                    raise BoardError('no candidate in detection area')

        finally:
            if debug_frame is not None:
                for detection in self.detections:
                    detection.debug_draw(debug_frame, color_bg=(0, 255, 255),
                            color_fg=(0, 0, 0))

        perf_ms = (time.perf_counter() - start_time) * 1000
#       logger.debug(f'get_programme() took {perf_ms:.3f} ms')

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

# TODO rewrite in concurrent.futures

import asyncio
import enum
import sys
import threading

import bleak

from loguru import logger

from . import (
    engine,
)

NORDIC_UART_RX = '6e400002-b5a3-f393-e0a9-e50e24dcca9e'
NORDIC_UART_TX = '6e400003-b5a3-f393-e0a9-e50e24dcca9e'

class Command(bytes, enum.Enum):
    TURN_LEFT =     b'H'
    BACKWARD =      b'J'
    FORWARD =       b'K'
    TURN_RIGHT =    b'L'
    SONAR =         b'S'


class BluetoothBot:
    def __init__(self, mac):
        self.mac = mac
        self.loop = None
        self.queue = None
        self.thread = None
        self.connected = False
        self.last_range = None

        self.engine = engine.RemoteControlEngine(bot=self)
        self.current_task = None

    def connect(self):
        assert self.loop is None
        self.thread = threading.Thread(target=asyncio.run, args=(self._main(),))
        self.thread.start()

    def disconnect(self):
        assert self.loop is not None
        asyncio.run_coroutine_threadsafe(
            self.queue.put((None, None)), self.loop)
        self.thread.join()

    def execute_programme(self, programme):
        logger.debug(f'{type(self).__name__}.execute_programme()')
        assert self.loop is not None
        assert self.current_task is None
        self.current_task = asyncio.run_coroutine_threadsafe(self._execute(programme), self.loop)

    def stop_programme(self):
        assert self.loop is not None
        assert self.current_task is not None
        self.current_task.cancel()


    async def _execute(self, programme):
        logger.debug(f'{type(self).__name__}._execute()')
        try:
            await programme.execute(self.engine)
        finally:
            self.current_task = None


    async def rpc(self, command):
        logger.debug(f'{type(self).__name__}.rpc({command=})')
        future = self.loop.create_future()
        await self.queue.put((command, future))
        return await future

    async def turn_left(self):
        await self.rpc(Command.TURN_LEFT)
    async def backward(self):
        await self.rpc(Command.BACKWARD)
    async def forward(self):
        await self.rpc(Command.FORWARD)
    async def turn_right(self):
        await self.rpc(Command.TURN_RIGHT)
    async def sonar(self):
        return (await self.rpc(Command.SONAR))[0]

    def rpc_threadsafe(self, command, *, timeout=None):
        assert self.loop is not None
        future = asyncio.run_coroutine_threadsafe(self.rpc(command), self.loop)
        return future.result(timeout=timeout)

    def turn_left_threadsafe(self):
        self.rpc_threadsafe(Command.TURN_LEFT)
    def backward_threadsafe(self):
        self.rpc_threadsafe(Command.BACKWARD)
    def forward_threadsafe(self):
        self.rpc_threadsafe(Command.FORWARD)
    def turn_right_threadsafe(self):
        self.rpc_threadsafe(Command.TURN_RIGHT)
    def sonar_threadsafe(self):
        return self.rpc_threadsafe(Command.SONAR)[0]


    def rpc_nowait(self, command):
        asyncio.run_coroutine_threadsafe(self.rpc(command), self.loop)

    def turn_left_nowait(self):
        self.rpc_nowait(Command.TURN_LEFT)
    def backward_nowait(self):
        self.rpc_nowait(Command.BACKWARD)
    def forward_nowait(self):
        self.rpc_nowait(Command.FORWARD)
    def turn_right_nowait(self):
        self.rpc_nowait(Command.TURN_RIGHT)
    def sonar_nowait(self):
        self.rpc_nowait(Command.SONAR)


    async def _main(self):
        self.loop = asyncio.get_running_loop()
        self.last_range = None

        try:
            async with bleak.BleakClient(self.mac) as client:
                nordic_uart_tx = client.services.get_characteristic(NORDIC_UART_TX)
                nordic_uart_rx = client.services.get_characteristic(NORDIC_UART_RX)

                self.connected = True
                self.queue = asyncio.Queue(1)
                while True:
                    command, future = await self.queue.get()
                    if command is None:
                        # disconnect
                        break

                    async def handle_tx(_characteristic, data):
                        logger.debug(f'handle_tx(..., {data=})')
                        future.set_result(data)
                        if command is Command.SONAR:
                            self.last_range = data[0]

                    await client.start_notify(nordic_uart_tx, handle_tx)

                    await client.write_gatt_char(nordic_uart_rx, command)
                    try:
                        await future
                    except asyncio.CancelledError:
                        pass
                    await client.stop_notify(nordic_uart_tx)
                    self.queue.task_done()

        except bleak.BleakError:
            # TODO register handler for disconnect
            pass

        finally:
            self.connected = False
            self.loop = None




# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import operator
from dataclasses import dataclass

from funcparserlib import parser as _parser

from . import (
    ast,
    blocks,
)

@dataclass(frozen=True)
class ParseToken:
    token: blocks.Token
    line: int
    col: int

    @classmethod
    def stream_from_lines(cls, lines):
        for lineno, line in enumerate(lines, start=1):
            i = 1
            for token in line:
                yield cls(token, lineno, i)
                i += 1
            yield cls(blocks.Token.NEWLINE, lineno, i)

def _line_cols(*parsed):
    lineset = {t.line for t in parsed if t is not None}
    line = lineset.pop()
    assert not lineset

    colset = set()
    for t in parsed:
        if isinstance(t, ast.Parsed):
            colset.update(t.cols)
        elif isinstance(t, ParseToken):
            colset.add(t.col)
        else:
            assert t is None
    return line, (min(colset), max(colset))

# make_ functions

def make_Number(parsed):
    sign_minus, digits = parsed
    return ast.Number(*_line_cols(sign_minus, *digits),
        int(''.join(d.token.text for d in digits))
        * (-1 if sign_minus is not None else 1))

def make_Variable(parsed):
    return ast.Variable(*_line_cols(parsed), parsed.token.text)

def make_Forever(parsed):
    return ast.Forever(*_line_cols(parsed))

def make_Assign(parsed):
    start, var, exp = parsed
    return ast.Assign(*_line_cols(*parsed), var.token.text, exp)

_ops = {
    blocks.Token.PLUS: operator.add,
    blocks.Token.MINUS: operator.sub,
}
def make_BinOp(parsed):
    first, rest = parsed
    left = first
    for o, right in rest:
        left = ast.BinOp(_line_cols(left, o, *right), _ops[o], left, right)
    return left

def make_Condition(parsed):
    d, o = parsed
    return ast.Condition(*_line_cols(*parsed),
        ast.Direction(d.token.text), ast.Object(o.token.text))

def make_Function(parsed):
    start, funcname, funcbody, end = parsed
    return ast.Function(*cls._line_cols(start, funcname), (start.line, end.line),
        funcbody, funcname.token.value)

def make_Repeat(parsed):
    start, counter, body, end = parsed
    return ast.Repeat(*_line_cols(start, counter), (start.line, end.line),
        body, counter)

def make_If(parsed):
    start, condition, body, orelse = parsed
    if isinstance(orelse, ParseToken) and orelse.token == blocks.Token.END_BLOCK:
        lines = (start.line, orelse.line)
        orelse = None
    else: # If or Else
        lines = (start.line, orelse.lines[1])
    return ast.If(*_line_cols(start, condition), lines,
        body, condition, orelse)

def make_RepeatWhile(parsed):
    start, cond, body, end = parsed
    return ast.RepeatWhile(*_line_cols(start, cond), (start.line, end.line),
        body, cond)

def make_simple(cls):
    def make_simple(parsed):
        if isinstance(parsed, ParseToken):
            parsed = [parsed]
        return cls(*_line_cols(*parsed), *parsed[1:])
    return make_simple

def make_CallFunction(parsed):
    start, funcname = parsed
    return ast.CallFunction(*_line_cols(*parsed), funcname.token.text)

def make_block(cls):
    def make_block(parsed):
        start, body, end = parsed
        return cls(*_line_cols(start), (start.line, end.line), body)
    return make_block


# helper parsers

def tok(token):
    return _parser.some(lambda t: t.token == token).named(token.name)

def toktype(type):
    return _parser.some(lambda t: t.token.type == type).named(type)

def toktype_(type):
    return _parser.skip(toktype(type))

def name(parsed):
    return parsed.token.value

# grammar

newline = toktype_(blocks.TokenType.NEWLINE)

digit = toktype(blocks.TokenType.DIGIT)
varname = toktype(blocks.TokenType.VARNAME)
funcname = toktype(blocks.TokenType.FUNCNAME)
op = toktype(blocks.TokenType.OP)
direction = toktype(blocks.TokenType.DIRECTION)
obj = toktype(blocks.TokenType.OBJECT)

minus = tok(blocks.Token.MINUS)
equals = _parser.skip(tok(blocks.Token.EQUALS))

number = (_parser.maybe(minus) + _parser.oneplus(digit)
    ).named('number') >> make_Number
term = number | varname
expr = (term + _parser.many(op + term)).named('expr') >> make_BinOp
forever = tok(blocks.Token.FOREVER) >> make_Forever

condition = (direction + obj).named('condition') >> make_Condition

statement = _parser.forward_decl().named('statement')
body = _parser.many(statement).named('body')
end_block = tok(blocks.Token.END_BLOCK)

simple_stmt = (
    tok(blocks.Token.TURN_LEFT) >>      make_simple(ast.TurnLeft) |
    tok(blocks.Token.TURN_RIGHT) >>     make_simple(ast.TurnRight) |
    tok(blocks.Token.JUMP) >>           make_simple(ast.Jump) |
    tok(blocks.Token.PICK_UP) >>        make_simple(ast.PickUp) |
    tok(blocks.Token.ACTIVATE) >>       make_simple(ast.Activate) |
    tok(blocks.Token.DRAW) >>           make_simple(ast.Draw) |
    tok(blocks.Token.BREAK) >>          make_simple(ast.Break) |
    tok(blocks.Token.BREAK_FUNCTION) >> make_simple(ast.BreakFunction) |

    (tok(blocks.Token.STEP) + expr)
        >> make_simple(ast.Step) |
    (tok(blocks.Token.PLACE) + direction)
        >> make_simple(ast.Place) |
    (tok(blocks.Token.VARIABLE) + varname + equals + expr)
        >> make_simple(ast.Assign) |

    (tok(blocks.Token.CALL_FUNCTION) + funcname)
        >> make_CallFunction
) + newline

repeat_stmt = (
    tok(blocks.Token.REPEAT) + (forever | expr) + newline +
    body +
    end_block + newline
) >> make_Repeat

repeat_while_stmt = (
    tok(blocks.Token.REPEAT_WHILE) + condition + newline +
    body +
    end_block + newline
) >> make_RepeatWhile

else_stmt = (
    tok(blocks.Token.ELSE) + newline +
    body +
    end_block + newline
) >> make_block(ast.Else)

elif_stmt = _parser.forward_decl()
elif_stmt.define(
    tok(blocks.Token.ELSE_IF) + condition + newline +
    body +
    (elif_stmt >> make_If | else_stmt | end_block + newline)
)

if_stmt = (
    tok(blocks.Token.IF) + condition + newline +
    body +
    (elif_stmt >> make_If | else_stmt | end_block + newline)
) >> make_If

statement.define(
    simple_stmt |
    repeat_stmt |
    repeat_while_stmt |
    if_stmt
)

programme = (
    (tok(blocks.Token.BEGIN) + newline).named('begin') +
    body +
    (tok(blocks.Token.END) + newline).named('end') +
    _parser.skip(_parser.finished)
) >> make_block(ast.Programme)

function = (
    (tok(blocks.Token.DEFINE_FUNCTION) + funcname + newline).named('define') +
    body +
    (tok(blocks.Token.END_FUNCTION) + newline).named('endfunc') +
    _parser.skip(_parser.finished)
)

parser = programme | function

def parse(lines):
    return parser.parse(list(ParseToken.stream_from_lines(lines)))

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import pprint

from tuxgo import (
    blocks,
    parser,
)

def test_parser():
    parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.STEP, blocks.Token.DIGIT_6),
            (blocks.Token.REPEAT, blocks.Token.DIGIT_3),
                (blocks.Token.TURN_RIGHT,),
                (blocks.Token.STEP, blocks.Token.DIGIT_2),
                (blocks.Token.IF, blocks.Token.HERE, blocks.Token.COLLECTABLE),
                    (blocks.Token.PICK_UP,),
                (blocks.Token.ELSE_IF, blocks.Token.IN_FRONT, blocks.Token.MARK_1),
                    (blocks.Token.BREAK,),
                (blocks.Token.ELSE,),
                    (blocks.Token.CALL_FUNCTION, blocks.Token.A),
                (blocks.Token.END_BLOCK,),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

def test_if():
    parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

def test_if_else():
    parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.ELSE,),
                (blocks.Token.STEP, blocks.Token.DIGIT_2),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

def test_if_elif():
    parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.ELSE_IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_2),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

def test_if_elif_else():
    parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.ELSE_IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_2),
            (blocks.Token.ELSE,),
                (blocks.Token.STEP, blocks.Token.DIGIT_3),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

def test_if_elif_elif():
    pprint.pprint(parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.ELSE_IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_2),
            (blocks.Token.ELSE_IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_3),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ]))

def test_if_elif_elif_else():
    pprint.pprint(parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.ELSE_IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_2),
            (blocks.Token.ELSE_IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_3),
            (blocks.Token.ELSE,),
                (blocks.Token.STEP, blocks.Token.DIGIT_4),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ]))

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import dataclasses
import enum
import tomllib
import typing

from dataclasses import dataclass
from typing import (
    Optional,
)

from . import (
    assets,
    ast,
)


class TokenType(str, enum.Enum):
    BEGIN = enum.auto()
    END = enum.auto()

    SIMPLE_STMT = enum.auto()
    DIR_STMT = enum.auto()
    EXPR_STMT = enum.auto()

    DIGIT = enum.auto()
    DIRECTION = enum.auto()
    OP = enum.auto()

    FUNCNAME = enum.auto()
    VARNAME = enum.auto()
    OBJECT = enum.auto()

    NEWLINE = enum.auto()

    def __repr__(self):
        return f'{type(self).__name__}.{self.name}'


@dataclass(frozen=True)
class TokenDataMixIn:
    text: str
    type: Optional[TokenType] = dataclasses.field(default=None)

class Token(TokenDataMixIn, enum.Enum):
    BEGIN =             'START', TokenType.BEGIN
    END =               'KONIEC', TokenType.END

    TURN_LEFT =         'OBRÓT W LEWO', TokenType.SIMPLE_STMT
    TURN_RIGHT =        'OBRÓT W PRAWO', TokenType.SIMPLE_STMT
    JUMP =              'SKOCZ', TokenType.SIMPLE_STMT
    PICK_UP =           'PODNIEŚ', TokenType.SIMPLE_STMT
    ACTIVATE =          'UŻYJ', TokenType.SIMPLE_STMT
    DRAW =              'RYSUJ', TokenType.SIMPLE_STMT

    STEP =              'KROK', TokenType.EXPR_STMT
    PLACE =             'UMIEŚĆ', TokenType.DIR_STMT
    VARIABLE =          'ZMIENNA',

    REPEAT =            'POWTÓRZ',
    REPEAT_WHILE =      'POWTÓRZ DOPÓKI',
    BREAK =             'PRZERWIJ',
    IF =                'JEŻELI',
    ELSE_IF =           'W PRZECIWNYM RAZIE JEŻELI',
    ELSE =              'W PRZECIWNYM RAZIE',
    END_BLOCK =         'KONIEC PĘTLI LUB WARUNKU',

    FOREVER =           'CIĄGLE',

    HERE =              ast.Direction.HERE.value, TokenType.DIRECTION
    IN_FRONT =          ast.Direction.IN_FRONT.value, TokenType.DIRECTION
    ON_THE_LEFT =       ast.Direction.ON_THE_LEFT.value, TokenType.DIRECTION
    ON_THE_RIGHT =      ast.Direction.ON_THE_RIGHT.value, TokenType.DIRECTION
    BEHIND =            ast.Direction.BEHIND.value, TokenType.DIRECTION

    X =                 'X', TokenType.VARNAME
    Y =                 'Y', TokenType.VARNAME
    EQUALS =            '=',
    PLUS =              '+', TokenType.OP
    MINUS =             '-', TokenType.OP

    SWITCH_TO_HERO =    'STERUJ (SCOTTIE)', TokenType.SIMPLE_STMT
    SWITCH_TO_RED =     'STERUJ (CZERWONY BÓBR)', TokenType.SIMPLE_STMT
    SWITCH_TO_BLUE =    'STERUJ (NIEBIESKI BÓBR)', TokenType.SIMPLE_STMT
    SWITCH_TO_YELLOW =  'STERUJ (ŻÓŁTY BÓBR)', TokenType.SIMPLE_STMT

    CALL_FUNCTION =     'WYWOŁAJ FUNKCJĘ',
    DEFINE_FUNCTION =   'DEFINIUJ FUNKCJĘ', TokenType.BEGIN
    END_FUNCTION =      'KONIEC FUNKCJI', TokenType.END
    BREAK_FUNCTION =    'PRZERWIJ FUNKCJĘ',
    A =                 'A', TokenType.FUNCNAME
    B =                 'B', TokenType.FUNCNAME
    C =                 'C', TokenType.FUNCNAME

    OBSTACLE =          ast.Object.OBSTACLE.value, TokenType.OBJECT
    ELEVATED =          ast.Object.ELEVATED.value, TokenType.OBJECT
    MARK_1 =            ast.Object.MARK_1.value, TokenType.OBJECT
    MARK_2 =            ast.Object.MARK_2.value, TokenType.OBJECT
    COLLECTABLE =       ast.Object.COLLECTABLE.value, TokenType.OBJECT
    ACTION =            ast.Object.ACTION.value, TokenType.OBJECT
    EMPTY =             ast.Object.EMPTY.value, TokenType.OBJECT

    DIGIT_0 =           '0', TokenType.DIGIT
    DIGIT_1 =           '1', TokenType.DIGIT
    DIGIT_2 =           '2', TokenType.DIGIT
    DIGIT_3 =           '3', TokenType.DIGIT
    DIGIT_4 =           '4', TokenType.DIGIT
    DIGIT_5 =           '5', TokenType.DIGIT
    DIGIT_6 =           '6', TokenType.DIGIT
    DIGIT_7 =           '7', TokenType.DIGIT
    DIGIT_8 =           '8', TokenType.DIGIT
    DIGIT_9 =           '9', TokenType.DIGIT

    NEWLINE =           None, TokenType.NEWLINE

    def __repr__(self):
        return f'{type(self).__name__}.{self.name}'

class Block(typing.NamedTuple):
    aruco_id: int
    id: int
    token: int

    def __str__(self):
        return f'{self.id:03d} {self.token}' if self.id is not None else f'--- {self.token}'


class BlockLoader:
    def __init__(self, path=assets / 'blocks.toml'):
        self._by_id = {}
        self._by_aruco_id = {}
        with open(path, 'rb') as file:
            data = tomllib.load(file)

        for tile in data['block']:
            tile.setdefault('id', None)
            tile['token'] = Token[tile['token']]
            tile = Block(**tile)
            self._by_aruco_id[tile.aruco_id] = tile
            if tile.id is not None:
                self._by_id[tile.id] = tile

    def get_block_by_aruco_id(self, key):
        return self._by_aruco_id[key]

    def get_block_by_id(self, key):
        return self._by_id[key]

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import os
import pathlib

import cv2
import pytest

from tuxgo import (
    blocks,
    detector,
)

@pytest.fixture
def load_frame():
    def load_frame(filename):
        return cv2.imread(os.fspath(pathlib.Path(__file__).parent / filename))
    return load_frame

@pytest.fixture
def detect_markers_in_frame(load_frame):
    def detect_markers_in_frame(frame):
        return detector.Detector().detect_markers(load_frame(frame))
    return detect_markers_in_frame

def test_trivial(detect_markers_in_frame):
    analyser = detector.Analyser(detect_markers_in_frame('test_trivial.jpg'))
    assert list(analyser.get_programme()) == [
        (blocks.Token.BEGIN,),
        (blocks.Token.END,),
    ]

rotate_programme = [
    (blocks.Token.BEGIN,),
    (blocks.Token.STEP, blocks.Token.DIGIT_6),
    (blocks.Token.REPEAT, blocks.Token.DIGIT_3),
    (blocks.Token.TURN_RIGHT,),
    (blocks.Token.STEP, blocks.Token.DIGIT_2),
    (blocks.Token.PICK_UP,),
    (blocks.Token.END_BLOCK,),
    (blocks.Token.END,),
]

@pytest.mark.parametrize('rot', [0, 45, 90])
def test_rotate(detect_markers_in_frame, rot):
    analyser = detector.Analyser(detect_markers_in_frame(
        f'test_rotate_{rot:02d}.jpg'))
    assert list(analyser.get_programme()) == rotate_programme

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import os
import pathlib
from typing import Union

import click
import cv2.aruco
import numpy as np


class Generator:
    def __init__(self, dictionary: Union[cv2.aruco.Dictionary, int, str]):
        if isinstance(dictionary, str):
            dictionary = getattr(cv2.aruco, dictionary)
        if isinstance(dictionary, int):
            dictionary = cv2.aruco.getPredefinedDictionary(dictionary)
        self.dictionary = dictionary

    def range(self) -> range:
        return range(len(self.dictionary.bytesList))

    def generate_bitmap(self, id_: int) -> np.ndarray:
        return self.dictionary.drawMarker(
            id_, self.dictionary.markerSize + 2, borderBits=1
        ).astype(bool)

    def generate(
        self, id_: int, *, size=360: int, inverted=True: bool
    ) -> np.ndarray:
        im = self.dictionary.drawMarker(id_, size)
        if inverted:
            im = 255 - im
        return im

    def generate_to_file(
        self, id_: int, filename, *, size=360: int, inverted=True: bool
    ):
        im = self.generate(id_, size=size, inverted=inverted)
        cv2.imwrite(os.fspath(filename), im)


@click.command()
@click.option('--invert/--no-invert', default=True)
@click.option('--size', type=int, default=360)
@click.option('--output-dir', '-o', type=pathlib.Path, default='.')
@click.argument('dictionary')
@click.argument('id_', type=int)
@click.pass_context
def main(ctx, invert, size, dictionary, id_, output_dir):
    try:
        generator = Generator(dictionary)
    except AttributeError:
        ctx.fail(f'no such dictionary: {dictionary!r}')
    
    inv = 'I' if invert else 'N'
    if id_ < 0:
        with click.progressbar(generator.range()) as bar:
            for id_ in bar:
                generator.generate_to_file(id_,
                    output_dir / f'{dictionary}-{id_:03}-{inv}.png',
                    size=size, inverted=invert)
    else:
        generator.generate_to_file(id_,
            output_dir / f'{dictionary}-{id_:03}-{inv}.png',
            size=size, inverted=invert)


if __name__ == '__main__':
    main()

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

from tuxgo import (
    ast,
    blocks,
    engine,
    parser,
)

class Engine(engine.ExecEngine):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.steps = 0
        self.turns = 0
        self.jumps = 0

    def step(self, counter: int):
        self.steps += counter

    def turn_left(self):
        self.turns += 1

    def turn_right(self):
        self.turns -= 1

    def jump(self):
        self.jumps += 1

    def pick_up(self):
        pass

    def place(self, direction: ast.Direction):
        pass

    def activate(self):
        pass

    def draw(self):
        pass

    def detect(self, direction: ast.Direction, obj: ast.Object):
        if direction == ast.Direction.IN_FRONT:
            if obj == ast.Object.OBSTACLE:
                return self.steps >= 10
            else:
                return self.steps < 10
        if direction == ast.Direction.HERE:
            return obj != ast.Object.EMPTY
        raise NotImplementedError()

def test_step():
    programme = parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.STEP, blocks.Token.DIGIT_6),
            (blocks.Token.STEP, blocks.Token.DIGIT_9),
        (blocks.Token.END,),
    ])

    engine = Engine()
    programme.execute(engine)
    assert engine.steps == 15

def test_turns():
    programme = parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.TURN_LEFT,),
            (blocks.Token.TURN_LEFT,),
            (blocks.Token.TURN_RIGHT,),
        (blocks.Token.END,),
    ])

    engine = Engine()
    programme.execute(engine)
    assert engine.turns == 1

def test_if():
    programme = parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.IF, blocks.Token.HERE, blocks.Token.EMPTY),
                (blocks.Token.TURN_LEFT,),
            (blocks.Token.ELSE,),
                (blocks.Token.TURN_RIGHT,),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

    engine = Engine()
    programme.execute(engine)
    assert engine.turns == -1

def test_repeat():
    programme = parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.REPEAT, blocks.Token.DIGIT_4, blocks.Token.DIGIT_2),
                (blocks.Token.TURN_LEFT,),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

    engine = Engine()
    programme.execute(engine)
    assert engine.turns == 42

def test_repeat_forever():
    programme = parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.REPEAT, blocks.Token.FOREVER),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
                (blocks.Token.IF, blocks.Token.IN_FRONT, blocks.Token.OBSTACLE),
                    (blocks.Token.BREAK,),
                (blocks.Token.END_BLOCK,),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

    engine = Engine()
    programme.execute(engine)
    assert engine.steps == 10

def test_repeat_while():
    programme = parser.parse([
        (blocks.Token.BEGIN,),
            (blocks.Token.REPEAT_WHILE, blocks.Token.IN_FRONT, blocks.Token.EMPTY),
                (blocks.Token.STEP, blocks.Token.DIGIT_1),
            (blocks.Token.END_BLOCK,),
        (blocks.Token.END,),
    ])

    engine = Engine()
    programme.execute(engine)
    assert engine.steps == 10

# vim: tw=80 ts=4 sts=4 sw=4 et



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: 2023 Wojtek Porczyk <woju@hackerspace.pl>

import abc
import dataclasses
import enum

from abc import abstractmethod
from dataclasses import dataclass
from typing import (
    Callable,
    List,
    Optional,
    Tuple,
    Union,
)

class Direction(enum.Enum):
    HERE =          'TUTAJ'
    IN_FRONT =      'Z PRZODU'
    ON_THE_LEFT =   'Z LEWEJ'
    ON_THE_RIGHT =  'Z PRAWEJ'
    BEHIND =        'Z TYŁU'

    def __repr__(self):
        return f'{type(self).__name__}.{self.name}'

class Object(enum.Enum):
    OBSTACLE =      'PRZESZKODA'
    ELEVATED =      'WZNIESIENIE'
    MARK_1 =        'KOPERTA: KWADRAT'
    MARK_2 =        'KOPERTA: KRZYŻYK'
    COLLECTABLE =   'PRZEDMIOT DO ZEBRANIA'
    ACTION =        'POLE AKCJI'
    EMPTY =         'PUSTE POLE'

    def __repr__(self):
        return f'{type(self).__name__}.{self.name}'

@dataclass(frozen=True)
class Parsed(metaclass=abc.ABCMeta):
    line: int = dataclasses.field(repr=False)
    cols: Tuple[int, int] = dataclasses.field(repr=False)

class Expression(Parsed):
    @abstractmethod
    def eval(self, engine) -> int:
        raise NotImplementedError()

class Forever(Parsed):
    pass

@dataclass(frozen=True)
class Number(Expression):
    value: int

    def eval(self, engine):
        return engine.eval_Number(self)

@dataclass(frozen=True)
class Variable(Expression):
    name: str

    def eval(self, engine):
        return engine.eval_Variable(self)

@dataclass(frozen=True)
class BinOp(Expression):
    op: Callable[[int, int], int]
    left: Expression
    right: Expression

    def eval(self, engine):
        return engine.eval_BinOp(self)

@dataclass(frozen=True)
class Condition(Parsed):
    direction: Direction
    object: Object

    def eval(self, engine):
        return engine.eval_Condition(self)

@dataclass(frozen=True)
class Statement(Parsed):
    #: Range of lines. For simple statement this is (self.line, self.line).
    lines: Tuple[int, int] = dataclasses.field(init=False, repr=False)
    def __post_init__(self):
        # in case a child class overrode this attribute, we don't overwrite
        try:
            self.lines
        except AttributeError:
            # we're dataclass(frozen=True), so we need to resort to tricks
            object.__setattr__(self, 'lines', (self.line, self.line))
        
    @abstractmethod
    def execute(self, engine: 'BaseEngine'):
        raise NotImplementedError()

class TurnLeft(Statement):
    def execute(self, engine):
        return engine.execute_TurnLeft(self)

class TurnRight(Statement):
    def execute(self, engine):
        return engine.execute_TurnRight(self)

class Jump(Statement):
    def execute(self, engine):
        return engine.execute_Jump(self)

class PickUp(Statement):
    def execute(self, engine):
        return engine.execute_PickUp(self)

class Activate(Statement):
    def execute(self, engine):
        return engine.execute_Activate(self)

@dataclass(frozen=True)
class Draw(Statement):
    def execute(self, engine):
        return engine.execute_Draw(self)

@dataclass(frozen=True)
class Step(Statement):
    counter: Expression

    def execute(self, engine):
        return engine.execute_Step(self)

@dataclass(frozen=True)
class Place(Statement):
    direction: Direction

    def execute(self, engine):
        return engine.execute_Place(self)

@dataclass(frozen=True)
class Assign(Statement):
    name: str
    expression: Expression

    def execute(self, engine):
        return engine.execute_Assign(self)

@dataclass(frozen=True)
class CallFunction(Statement):
    name: str

    def execute(self, engine):
        return engine.execute_CallFunction(self)

@dataclass(frozen=True)
class Break(Statement):
    def execute(self, engine):
        return engine.execute_Break(self)

@dataclass(frozen=True)
class BreakFunction(Statement):
    def execute(self, engine):
        return engine.execute_BreakFunction(self)

@dataclass(frozen=True)
class Block(Statement):
    lines: Tuple[int, int] = dataclasses.field(repr=False)
    body: List[Statement]
    body_lines: Tuple[int, int] = dataclasses.field(init=False, repr=False)

    def __post_init__(self):
        try:
            self.body_lines
        except AttributeError:
            object.__setattr__(self, 'body_lines', (
                min(stmt.lines[0] for stmt in self.body),
                max(stmt.lines[1] for stmt in self.body),
            ))

@dataclass(frozen=True)
class Repeat(Block):
    counter: Union[Expression, Forever]

    def execute(self, engine):
    	return engine.execute_Repeat(self)

@dataclass(frozen=True)
class ConditionStatement(Block):
    condition: Condition

@dataclass(frozen=True)
class RepeatWhile(ConditionStatement):
    def execute(self, engine):
        return engine.execute_RepeatWhile(self)

class Else(Block):
    def execute(self, engine):
        return engine.execute_Else(self)

@dataclass(frozen=True)
class If(ConditionStatement):
    orelse: Union['If', Else]
    def execute(self, engine):
        return engine.execute_If(self)

@dataclass(frozen=True)
class Programme(Block):
    def execute(self, engine):
        return engine.execute_Programme(self)

@dataclass(frozen=True)
class Function(Block):
    name: str
    def execute(self, engine):
        return engine.execute_Function(self)

# vim: tw=80 ts=4 sts=4 sw=4 et
